from radicl.probe import RAD_Probe
import pandas as pd
import numpy as np
from radicl.gps import USBGPS
from pynmeagps.nmeamessage import NMEAMessage
from pynmeagps.nmeahelpers import get_parts
from types import SimpleNamespace


def probe_not_connected():

    # Create a bool to use for skipping tests
    try:
        p = RAD_Probe()
        p.connect()
        not_connected = False

    except:
        not_connected = True

    return not_connected


def gps_not_connected():

    dev = USBGPS()
    if dev.cnx is None:
        not_connected = True
    else:
        not_connected = False

    return not_connected


class MockSerialPort:
    def __init__(self, **kwargs):
        # Add a byte array to be returned from read
        self.payload = 10

    def setDTR(self, num):
        pass

    def close(self):
        pass

    def reset_output_buffer(self):
        pass

    def write(self, data):
        return len(data)

    def read(self):
        pass

    def inWaiting(self):
        pass


class MockRADPort:
    def __init__(self, payload):
        self.payload = payload

    def openPort(self):
        pass

    def closePort(self):
        pass

    def flushPort(self):
        pass

    def writePort(self, data):
        return 1

    def writePortClean(self):
        return 1

    def readPort(self, nbytes):
        return self.payload

    def numBytesInBuffer(self):
        return len(self.payload)


class MockProbe:
    pass


class MOCKCLI:

    def grab_data(self, data_name):
        """
        Return a dataframe of the requested data
        Args:
            data_name: String of the data name to dataframe

        Returns:
            result: Dataframe
        """
        data = {}
        samples = 20
        if data_name == 'rawsensor':
            data['time'] = np.linspace(0, 1, samples)

            for c in ['Sensor1', 'Sensor2', 'Sensor3']:
                data[c] = np.linspace(1000, 4000, samples)

        elif data_name == 'filtereddepth':
            alt_samples = int(samples/3)
            data['time'] = np.linspace(0, 1, alt_samples)
            data['filtereddepth'] = np.linspace(100, 0, alt_samples)

        elif data_name == 'rawacceleration':
            alt_samples = int(samples/2)
            data['time'] = np.linspace(0, 1, alt_samples)
            data['Y-Axis'] = np.linspace(0, 2, alt_samples)

        result = pd.DataFrame(data)

        return result


class MockGPSStream:
    def __init__(self, payload):
        self.payload = payload

    def read(self):
        if self.payload:
            message = self.payload.pop()
            content, talker, msgid, payload, checksum = get_parts(message)

            return message,  NMEAMessage(
                    talker, msgid, 0, payload=payload, checksum=checksum)

        else:
            return b'', SimpleNamespace(msgID='NAN')




import pytest

from radicl.com import RAD_Serial
from radicl.api import RAD_API
from radicl.probe import RAD_Probe
from . import MOCKCLI


@pytest.fixture(scope='session')
def real_port():
    port = RAD_Serial(debug=True)
    port.openPort()
    yield port
    port.closePort()


@pytest.fixture(scope='session')
def real_api(real_port):
    yield RAD_API(real_port, debug=True)


@pytest.fixture(scope='session')
def real_probe():
    """
    Probe object with a reset at the end
    """
    prb = RAD_Probe(debug=True)
    prb.connect()

    yield prb
    prb.resetMeasurement()


@pytest.fixture()
def mock_cli():
    yield MOCKCLI()



import inspect
import unittest

import pytest

from radicl.probe import RAD_Probe
from radicl.ui_tools import *


@unittest.skip('Incomplete test. Needs work')
class TestUITools(unittest.TestCase):
    def test_parse_func_list(self):
        """
        Test we can parse the functions in the API for user interaction
        """
        api_funcs = inspect.getmembers(
            RAD_Probe, )  # predicate=inspect.ismethod)
        # print(api_funcs)
        result = parse_func_list(
            api_funcs, [
                'read', 'Data'], ignore_keywords=['correlation'])

        for v in ['depth', 'filtereddepth', 'rawpressure']:
            assert v in result.keys()

@pytest.mark.parametrize("idx, ratio, n_samples, expected", [
    (10, 1.1, 10, 9),
    (1, 0.9, 10, 0)
])
def test_get_index_from_ratio(idx, ratio, n_samples, expected):
    result = get_index_from_ratio(idx, ratio, n_samples)
    assert result == expected


if __name__ == '__main__':
    unittest.main()



# coding: utf-8

import datetime
import inspect
import struct
import time
import numpy as np
import pandas as pd

from . import __version__
from .com import RAD_Serial, find_kw_port
from .api import RAD_API
from .ui_tools import get_logger, parse_func_list
from .info import ProbeErrors, ProbeState, AccelerometerRange, SensorReadInfo


class RAD_Probe:
    """
    Class for directly interacting with the probe.

    """
    __data_buffer_guide = ['Raw Sensor', 'Acceleration', ' Raw Pressure',
                           'Raw Depth', 'Filtered Depth',
                           'Acceleration Correlation',
                           'Pressure/Depth Correlation',
                           'Depth Corrected Sensor']

    def __init__(self, ext_api: RAD_API=None, debug=False):
        """
        Args:
            ext_api: rad_api.RAD_API object pre-instantiated
        """

        self._state = ProbeState.NOT_SET
        self._last_state = ProbeState.NOT_SET
        self._sampling_rate = None
        self._accelerometer_range = None
        self._zpfo = None
        self._serial_number = None

        self._settings = None
        self._getters = None

        self.debug = debug
        self.api:RAD_API = ext_api
        self.available_devices = []

        self.log = get_logger(__name__, debug=self.debug)

    def update_devices(self):
        self.log.info("Scanning for COM ports...")
        self._available_devices = find_kw_port(['STMicroelectronics', 'STM32'])

    @property
    def serial_number(self):
        if self._serial_number is None:
            self._serial_number = self.getProbeSerial()
        return self._serial_number

    @property
    def multiple_devices_available(self):
        return len(self.available_devices) > 1

    @property
    def no_devices_available(self):
        return len(self.available_devices) == 0

    def connect(self, device=None):
        """ Attempt to establish a connection with the probe"""

        if self.api is None:
            # No external API object was provided. Create new serial and API
            # objects for internal use
            port = RAD_Serial(debug=self.debug)
            port.openPort()

            if not port:
                self.log.info("No device present")
            else:
                port.flushPort()
                # Create the API and FMTR instances The API class is linked to
                # the port object
                api = RAD_API(port, debug=self.debug)

                # Switch the device over to API mode
                api.sendApiPortEnable()
                self.api = api
                self.api.Identify()

        ret = self.getProbeMeasState()
        connected = True if ret is not None else False
        if not connected:
            self.log.error("Unable to connect to the probe. Unplug and"
                           " power cycle it.")
        return connected

    def disconnect(self):
        self.log.info("Disconnecting probe.")
        self.api.port.closePort()
        self.api = None 
        
    @property
    def state(self):
        return  self._state

    @property
    def last_state(self):
        return self._last_state

    @property
    def sampling_rate(self):
        if self._sampling_rate is None:
            self._sampling_rate = self.getSetting(setting_name='samplingrate')
        return self._sampling_rate

    @property
    def zpfo(self):
        if self._zpfo is None:
            self._zpfo = self.getSetting(setting_name='zpfo')
        return self._zpfo

    @property
    def accelerometer_range(self):
        # Grab the range to scale the incoming data
        if self._accelerometer_range is None:
            sensing_range = self.getSetting(setting_name='accrange')
            # Add in a default
            if sensing_range is None:
                sensing_range = 16
            self._accelerometer_range = sensing_range
        return self._accelerometer_range

    def _assign_settings_functions(self):
        # Manages the settings
        settings_funcs = inspect.getmembers(self.api,
                                            predicate=inspect.ismethod)
        ignores = ['reset']
        self._settings = parse_func_list(settings_funcs,
                                        ['Meas', 'Set'],
                                        ignore_keywords=ignores)
        self._getters = parse_func_list(settings_funcs,
                                       ['Meas', 'Get'],
                                       ignore_keywords=ignores)

    @property
    def settings(self):
        """Dictionary of function to set settings"""
        if self._settings is None:
            self._assign_settings_functions()
        return self._settings

    @property
    def getters(self):
        """ Dictionary of functions to get settings"""
        if self._getters is None:
            self._assign_settings_functions()
        return self._getters


    def manage_error(self, ret_dict, stack_id=1):
        """
        Handles the common scenario of looking at the returned Dictionary
        from the probe where there may be an error or simply a com error.
        This function reports the name of the function and reports the error
        function.

        Args:
            ret_dict: Dictionary of keys ['Status','data','errorCode']
            stack_id: number of functions up the stack to use for reporting
                      function name when errors occur.
        """
        name = inspect.stack()[stack_id][3]

        if ret_dict['errorCode'] is not None:
            self.log.error("{} error:{}".format(name, ret_dict['errorCode']))

        else:
            self.log.error("{} error: COM".format(name))

    def manage_data_return(self, ret, num_values=1, dtype=int):
        """
        Manage a return from the probe when were expecting an integer

        Args:
            ret: Dictionary of the return with keys status, errorCode, and data
            num_values: Number of expect values to be returned
            dtype: Data type expected back

        Returns:
            result: None when no data was received. Integer when its not.
        """
        result = []

        if (ret['status'] == 1) and (ret['data'] is not None):

            data_size = len(ret['data'])
            increment = int(data_size / num_values)

            if data_size % num_values != 0:
                raise IOError("Data received (length = {}) is not evenly "
                              "divided by the expected number of values = "
                              "{}".format(data_size, num_values))

            for i in range(0, data_size, increment):
                data = ret['data'][i:i + increment]
                if dtype == int:
                    value = int.from_bytes(data, byteorder='little')

                elif dtype == str:
                    value = data.decode('utf-8')

                elif dtype == 'hex':
                    value = data.hex()

                else:
                    raise ValueError(
                        "No types other than str, hex, and int are implemented")

                result.append(value)

            if num_values == 1:
                result = result[0]

        else:
            self.manage_error(ret, stack_id=2)
            result = None

        return result

    def readData_by_segment(self, buffer_id, segment):
        """
        Read segments one at a time
        """
        # Data Segments to collect
        result = False

        # Request the data
        ret = self.api.MeasReadDataSegment(buffer_id, segment)

        if ret['status'] == 1 and ret['data'] is not None:
            data_chunk = ret['data']
        else:
            data_chunk = None
        return data_chunk

    def get_number_of_segments(self, buffer_id):
        """ Retrieve the number of data segments available from measurement"""
        ret = self.api.MeasGetNumSegments(buffer_id)
        if ret['status'] == 1:
            num_segments = int.from_bytes(ret['data'], byteorder='little')
        else:
            num_segments = None
        return num_segments

    def __readData(self, buffer_id, max_retry=10, init_delay=0.004):
        """
        Private function to retrieve data from the probe.
         Args:
            buffer_id: Integer specifying location in the probe buffer
            max_retry: Integer number of attempts before exiting with a fail
        """

        result = False
        num_segments = 0
        data = []

        buffer_name = self.__data_buffer_guide[buffer_id]
        self.log.info("Querying probe for {} data...".format(buffer_name.lower()))

        # Final data to return
        final = {'status': 0, 'SegmentsAvailable': 0, 'SegmentsRead': 0,
                 'BytesRead': 0,
                 'data': None}

        # Get the number of data segments available
        num_segments = self.get_number_of_segments(buffer_id)
        if num_segments is not None:
            self.log.debug('Retrieving {:,} segments of {} data...'.format(num_segments, buffer_name.lower()))

        # No data returned
        else:

            self.log.debug(f"getNumSegments Error (buffer_id = {buffer_id})")

        # If we do have number of segments
        if num_segments != 0 and num_segments is not None:
            self.log.debug("Reading %d segments" % num_segments)
            byte_counter = 0

            # Data Segments to collect
            for ii in range(0, num_segments):
                result = False

                # initial delay time
                wait_time = init_delay

                # Delays and retry
                for jj in range(0, max_retry):

                    time.sleep(wait_time)

                    # Request the data
                    data_chunk = self.readData_by_segment(buffer_id, ii)

                    if data_chunk is not None:
                        data.extend(data_chunk)
                        byte_counter = len(data)
                        result = True
                        # Break the retry loop
                        break

                    else:
                        if jj >= max_retry:
                            self.log.warning('Missed data segment, after {0:d} attempts.'.format(jj + 1))

                        # Developer friendly response in event of read error
                        msg = ("{0} Data Error: Buffer ID = {1:d}, "
                               "Segment ID={2:d}/{3:d}, Retry #{4:d}, "
                               " COM Delay = {5}s "
                               "").format(buffer_name, buffer_id, ii,
                                          num_segments, jj, wait_time)

                        self.log.debug(msg)

                        # Increase the wait time every failed request
                        wait_time += wait_time

            # Was the data read successful?
            final['status'] = int(result)
            final['SegmentsAvailable'] = num_segments
            final['SegmentsRead'] = ii + 1  # Report this not zero based
            final['BytesRead'] = byte_counter

            if final['SegmentsRead'] > 0:
                final['data'] = data

        return final

    # ********************
    # * PUBLIC FUNCTIONS *
    # ********************
    def getProbeSerial(self):
        """
        Returns the probe's serial number. The return value is a string. If the
        request fails it will return None
        """

        ret = self.api.getSerialNumber()
        if ret['data'] is not None:
            # Flip the byte array since it comes in backwards
            ret['data'] = ret['data'][::-1]
            # upper case to match the store serial number lists
            result = self.manage_data_return(ret, dtype='hex') #.upper()
            if result is not None:
                result = result.upper()
        return result

    def getProbeSystemStatus(self):
        """
        Returns the probe's system status. The return value is an integer.
        If the request fails it will return None
        """

        ret = self.api.getSystemStatus()

        return self.manage_data_return(ret, dtype=int)

    def getProbeRunState(self):
        """
        Returns the probe's run state. The return value is an integer.
        If the request fails it will return None
        """

        ret = self.api.getRunState()
        return self.manage_data_return(ret, dtype=int)

    def getProbeMeasState(self):
        """
        Retrieves the probe measurement state and converts it to an integer

        Returns:
            integer-measurement state of the probe, or none if error arises.
        """
        data = None
        attempts = 0

        while data is None and attempts < 10:
            ret = self.api.getMeasState()
            data = self.manage_data_return(ret, dtype=int)
            attempts += 1

        if self.state != data:
            self._last_state = self._state
            self._state = ProbeState.from_state(data)

        return data

    def startMeasurement(self):
        """
        Starts a new measurement. Returns 1 if successful, 0 otherwise
        """

        ret = self.api.MeasStart()
        self.log.debug("Start measurement requested.")

        if ret['status'] == 1:
            self.wait_for_state(ProbeState.MEASURING)
            self.log.info("Measurement started...")

            return 1

        else:
            self.manage_error(ret)

            return 0

    def stopMeasurement(self):
        """
        Stops an ongoing measurement. Returns 1 if successful, 0 otherwise
        """

        ret = self.api.MeasStop()
        self.log.debug("Stop measurement requested.")

        if ret['status'] == 1:
            self.wait_for_state(ProbeState.PROCESSING)
            self.log.info("Measurement stopped...")

            return 1
        else:
            self.manage_error(ret)

            return 0

    def resetMeasurement(self):
        """
        Resets the measurement FSM to prepare for a new measurement.
        Returns 1 if successful, 0 otherwise
        """

        ret = self.api.MeasReset()
        self.log.debug("Measurement reset requested...")

        if ret['status'] == 1:
            result = self.wait_for_state(ProbeState.IDLE, delay=0.1)
            self.log.info("Probe measurement reset...")

            return 1

        else:
            self.manage_error(ret)

            return 0

    def wait_for_state(self, state:ProbeState, retry=500, delay=0.2):
        """
        Waits for the specified state to occur. This is particularly useful when
        a command is requested.

        Args:
            state: single integer
            retry: Number of attempts to try while Waiting for the states
            delay: time in seconds to wait between each attempt

        """

        attempts = 0
        result = False
        self.log.info(f"Waiting for state {state.value}, current state = {self.state.value}")

        while not result:
            result = self.state == state

            # Check for a probe advanced past the state
            if state != ProbeState.IDLE:
                if self.state >= state:
                    result = True

            if ProbeState.ready(state):
                if self.state == ProbeState.IDLE:
                        result = True

            if attempts > retry:
                self.log.error(
                    "Retry Exceeded waiting for state(s) {0}".format(state))
                result = False
                break
            else:
                attempts += 1

            time.sleep(delay)

            # Update the state
            self.getProbeMeasState()

        if result:
            self.log.debug(
                "{} queries while waiting for state {}".format(
                    attempts, state))

        return result

    def read_check_data_integrity(self, buffer_id, ret_dict, nbytes_per_value=None,
                                  nvalues=None, from_spi=False):
        """
        Receives a data function and  performs the data integrity check
        If the data is from _spi then we know how long the segments are.
        If there are not, then it is possible we receive an incomplete segment
        so we check for integer_multiples of that data

        Args:
            ret_dict: Returned dictionary of data and supporting meta
            nbytes_per_value: how many bytes expected per value per sample
            nvalues: Number of values expecter per sample
            from_spi: Is the data coming from SPI flash which guarantees a
                      segment size
        Returns:
            final: dict containing data in bytes, number of samples and
                   segments and return status.
        """

        buffer_name = self.__data_buffer_guide[buffer_id]
        final = None

        # successfully read data
        if ret_dict['data'] is None:
            self.log.error('Read error: No data available!')

        else:

            # ***** DATA INTEGRITY CHECK *****
            # Check for all segments read in
            all_segments = (ret_dict['SegmentsAvailable'] == \
                            ret_dict['SegmentsRead'])
            int_multiple = nbytes_per_value * nvalues

            # Grab the number of samples, NOTE this is only valid if the
            # ...conditions are true below
            samples = ret_dict['BytesRead'] // int_multiple

            # Data from SPI Flash
            if from_spi:
                expected_bytes = ret_dict['SegmentsRead'] * 256

                # Check we read all bytes:
                complete_bytes = expected_bytes == ret_dict['BytesRead']
                self.log.debug('Downloaded {:0,.2f}/{:0,.2f} Kb.'.format(
                    ret_dict['BytesRead'] / 1000,
                    expected_bytes / 1000))
            # From chip memory
            else:
                # We can have incomplete segments, so check for even numbers
                complete_bytes = ret_dict['BytesRead'] % int_multiple == 0
                self.log.debug('Byte Multiples: {:0,.2f}'.format(ret_dict['BytesRead'] / int_multiple))

            # Check the data integrity
            if not all_segments or not complete_bytes:
                self.log.error("Data Integrity Error: Unable to retrieve all "
                               "data for {}.".format(buffer_name))
                self.log.debug('All Segments not downloaded: {}, All bytes '
                               'downloaded: {}'.format(all_segments, complete_bytes))
            else:
                final = ret_dict
                final['samples'] = samples

                # Final reporting
                self.log.info('Retrieving {:,} samples of {} data...'
                              ''.format(ret_dict['samples'], buffer_name))

            self.log.debug("Segment Retrieved {:,d}/{:,d}."
                           " Bytes Retrieved {:0,.2f} Kb"
                           "".format(ret_dict['SegmentsRead'],
                                     ret_dict['SegmentsAvailable'],
                                     ret_dict['BytesRead'] / 1000))

            return final

    def unpack_sensor(self, data, sensor:SensorReadInfo):
        """
        Attempt to standardize the conversion of downloaded data for more usages
        Args:
            data: Data to unpack
            sensor: Sensor storage info
            extra_conversion: Option to add a more dynamic conversion like in accelerometer

        Returns:
            final: Dictionary of unpacked data

        """
        offset = 0
        final = {name: [] for name in sensor.data_names}

        unpack_bytes = sensor.unpack_type is not None
        convert = sensor.conversion_factor is not None

        samples = len(data) // sensor.bytes_per_sample

        for ii in range(0, samples):

            # Loop over each sensor and add it to the final dict
            for idx, name in enumerate(sensor.data_names):
                # Form the index for the bytes for the start of each value.
                byte_idx = idx * sensor.nbytes_per_value + offset

                # Certain datasets need to be unpacked
                if unpack_bytes:
                    # Form a byte object and unpack it
                    byte_list = data[byte_idx: (byte_idx + sensor.nbytes_per_value)]
                    byte_object = bytes(byte_list)
                    value = struct.unpack(sensor.unpack_type, byte_object)[0]

                else:
                    value = data[byte_idx] + \
                                   data[byte_idx + 1] * sensor.bytes_per_segment

                # Perform conversion of units
                if convert:
                    value *= sensor.conversion_factor

                # Grab each sensor value in the byte array
                final[name].append(value)

            # Move farther down the bytes to read the next segment.
            offset += sensor.nbytes_per_value * sensor.expected_values
        df = pd.DataFrame.from_dict(final)

        # Special Treatment of the accelerometer data
        if sensor == SensorReadInfo.ACCELEROMETER:
            self.log.info('Scaling accelerometer data')
            sensitivity = AccelerometerRange.from_range(self.accelerometer_range)
            df = df.mul(sensitivity.value_scaling)

        df = self.time_decimate(df, sensor)
        return df

    def time_decimate(self, df, sensor: SensorReadInfo):
        """
        Form the data into a dataframe and scale it according to the ratio of max
        sample rate as is done in the FW
        """
        n_samples = df.index.size
        # Decimation ratio for peripheral sensors
        ratio = self.sampling_rate / SensorReadInfo.RAWSENSOR.max_sample_rate
        sr = int(sensor.max_sample_rate * ratio)
        seconds = np.linspace(0, n_samples / sr, n_samples)
        df['time'] = seconds
        df = df.set_index('time')
        return df

    def _parse_data(self, sensor):
        ret = self.__readData(sensor.buffer_id)
        ret = self.read_check_data_integrity(sensor.buffer_id, ret, nbytes_per_value=sensor.nbytes_per_value,
                                             nvalues=sensor.expected_values, from_spi=sensor.uses_spi)
        final = None
        if ret is not None:
            final = self.unpack_sensor(ret['data'], sensor)
        return final

    def readRawSensorData(self):
        """
        Reads the RAW sensor data.
        helpme - Raw NIR data for 4 sensors

        Returns:
            dict - containing data or None if read failed
        """
        sensor = SensorReadInfo.RAWSENSOR
        return self._parse_data(sensor)

    # TODO update for calibration scheme
    # def readCalibratedSensorData(self):
    #     """
    #     Reads the Read Calibrated sensor data.
    #     helpme - Calibrated data for 4 sensors
    #
    #     Returns:
    #         dict - containing data or None if read failed
    #     """
    #     calib_data = {}
    #     raw = self.readRawSensorData()
    #
    #     for id in range(1, 5):
    #         sensor = "Sensor{}".format(id)
    #         d = self.getSetting(setting_name='calibdata', sensor=id)
    #
    #         # For the calibration for sensor 1 we invert
    #         if id == 1:
    #             # Set the slope to the negative difference
    #             m = 4095 / (d[0] - d[1])
    #             # Set the intercept to the LOW value
    #             b = d[1]
    #
    #         else:
    #             # Set the slope to the positive difference
    #             m = 4095 / (d[1] - d[0])
    #             # Set the intercept to the LOW value
    #             b = d[0]
    #
    #         calib_data[sensor] = [m * (x - b) for x in raw[sensor]]
    #
    #     return calib_data

    def readRawAccelerationData(self):
        """
        Reads the raw 3 axis  acceleration data
        helpme - Raw 3 axis accelerometer data from the probe

        Returns:
            dict: containing accel data (x,y,z) or None if read failed

        """
        sensor = SensorReadInfo.ACCELEROMETER
        return self._parse_data(sensor)

    def readRawPressureData(self):
        """
        Reads the RAW pressure data, including the correlation index
        """
        sensor = SensorReadInfo.RAW_BAROMETER_PRESSURE
        return self._parse_data(sensor)

    def readFilteredDepthData(self):
        """
        Retrieves the filtered depth data according to the zero-phase low-pass
        filter (equivalent to Matlab's 'filtfilt'). The amount of filtering
        applied is set by :func:`~radicl.RAD_API.MeasSetZPFO`

        helpme - The probe's filtered depth data is the raw depth data made
                 relative to the change and the amount filtered set by the ZPFO
                 option.

        """
        sensor = SensorReadInfo.FILTERED_BAROMETER_DEPTH
        return self._parse_data(sensor)

    def readMeasurementTemperature(self):
        """
        Reads the temperature from the last measurement
        """

        ret = self.api.MeasGetMeasTemp()
        result = struct.unpack('<i', ret['data'][5:])[0]
        return result

    def getProbeHeader(self):
        """
        Returns the lines that should be in every data file produced.
        """
        t = datetime.datetime.now()
        fstr = "{0}-{1:02d}-{2:02d}--{3:02d}:{4:02d}:{5:02d}"
        time_stamp = fstr.format(t.year, t.month, t.day,
                                 t.hour, t.minute, t.second)
        header = {"RECORDED": time_stamp,
                  "radicl VERSION": __version__,
                  "FIRMWARE REVISION": self.api.full_fw_rev,
                  "HARDWARE REVISION": self.api.hw_rev,
                  "MODEL NUMBER": self.api.hw_id,
                  "Serial Num.":self.serial_number,
                  "Baro Temp.":self.readMeasurementTemperature(),
                  "ACC. Range": str(self.accelerometer_range)}
        return header

    def getSetting(self, setting_name=None, sensor=None):
        """
        Reads the probes setting from the dictionary of functions. Calls the
        function and manages the data.

        Args:
            setting_name: name of the function minus Meas and get
            sensor: Sensor ID for calibration data

        Returns:
            int: from the function getting the probe setting, or list of 2 for calibration data
        """
        if setting_name == 'calibdata':
            ret = self.getters[setting_name](sensor)
            num_values = 2
        else:
            ret = self.getters[setting_name]()
            num_values = 1

        return self.manage_data_return(ret, num_values=num_values, dtype=int)

    def setSetting(self, setting_name=None, sensor=None, value=None, low_value=None,
                   hi_value=None):
        """
        sets the probe's setting
        """

        if setting_name == 'calibdata':
            ret = self.settings[setting_name](sensor, low_value, hi_value)

        else:
            ret = self.settings[setting_name](value)
        # Successful change!
        if ret['status'] == 1:
            # Reset properties to pull again
            if setting_name == 'accrange':
                self._accelerometer_range = None

            elif setting_name == 'samplingrate':
                self._sampling_rate = None

            elif setting_name == 'zpfo':
                self._zpfo = None

            return True

        else:
            self.manage_error(ret)

        return None



#!/usr/bin/env python

import argparse
import os
import sys
import platform
import time
import traceback

import numpy as np
from matplotlib import pyplot as plt
import matplotlib
import traceback
from radicl.ui_tools import get_logger, get_index_from_ratio
from study_lyte.profile import LyteProfileV6, Sensor
from study_lyte.styles import SensorStyle, EventStyle
from study_lyte.plotting import plot_events


matplotlib.rcParams['agg.path.chunksize'] = 100000


def plot_nir_depth_correct(ax, profile):
    ax.grid(True, axis='y', alpha=0.5)

    for sensor in ['Sensor2', 'Sensor3']:
        style = SensorStyle.from_column(sensor)
        ax.plot(profile.nir[sensor],
                profile.nir.depth,
                alpha=0.3, color=style.color, label=style.label)

    style = SensorStyle.ACTIVE_NIR
    ax.plot(profile.nir['nir'], profile.nir['depth'], color=style.color, label=style.label)
    ax.set_title("NIR Depth Corrected")
    ax.legend(loc='lower left', fontsize='small')
    return ax


def plot_force_depth_corrected(ax, profile, ):
    # plot the depth corrected Force
    force_style = SensorStyle.RAW_FORCE
    ax.grid(True, axis='y', which='both', alpha=0.5)
    ax.plot(profile.force['force'], profile.force['depth'], color=force_style.color, label=force_style.label)
    ax.set_title("Force Depth Corrected")
    ax.legend(loc='lower left', fontsize='small')
    ax.set_ylabel('Depth [cm]')
    ax.set_xlim((0, 4096))
    return ax

def report_profile_error(ax):
    ylims = ax.get_ylim()
    xlims = ax.get_xlim()
    ax.annotate('Failed to compute \ncorrected profile.', (np.mean(xlims)*0.25, np.mean(ylims)))
    return ax


def plot_hi_res(fname=None, timed_plot=None, calibration_dict={}):
    """
    Plots the timeseries, the depth corrected, accelerometer and depth data.
    Plot from a dataframe or from a file. Use auto close to auto close the figure
    after a some amount of time.

    Args:
        fname: Path to csv containing hi resolution data
        timed_plot: Amount of time to show the plot, if none user has to close it
        calibration_dict: Dictionary to offer calibration coefficients for any of the sensors

    """
    log = get_logger('Hi Res Plot')
    if 'Linux' in platform.platform():
        matplotlib.use('TkAgg')

    # Setup a panel of plots
    fig = plt.figure(figsize=(10, 6), constrained_layout=True)
    gs = fig.add_gridspec(2, 5)

    print('')
    if fname is not None:
        profile = LyteProfileV6(fname)
        log.info(f"Filename: {profile.filename}")
        plt.suptitle(os.path.basename(profile.filename))

    # print out some handy numbers
    log.info(profile.report_card())

    # Plot time series data force data
    ax = fig.add_subplot(gs[:, 0])
    plot_events(ax, profile.events, plot_type='vertical')
    force_style = SensorStyle.from_column('Sensor1')
    ax.plot(profile.raw['Sensor1'], profile.time, color=force_style.color)
    ax.set_title("Raw Force Timeseries")
    ax.legend(loc='lower left')
    ax.set_ylabel('Time [s]')
    ax.set_xlim(0, 4096)
    ax.invert_yaxis()

    # plot time series NIR data
    ax = fig.add_subplot(gs[:, 1])
    plot_events(ax, profile.events, plot_type='vertical')
    for sensor in ['Sensor2', 'Sensor3']:
        style = SensorStyle.from_column(sensor)
        ax.plot(profile.raw[sensor], profile.time, color=style.color, label=style.label)
    ax.set_title("NIR Timeseries")
    ax.legend(loc='lower left')
    ax.invert_yaxis()

    # plot the depth corrected Force
    ax = fig.add_subplot(gs[:, 2])
    try:
        ax = plot_force_depth_corrected(ax, profile)
    except:
        log.error(traceback.format_exc())
        report_profile_error(ax)

    # plot depth corrected NIR data
    ax = fig.add_subplot(gs[:, 3])
    try:
        ax = plot_nir_depth_correct(ax, profile)
    except:
        log.error(traceback.format_exc())
        report_profile_error(ax)

    #### plot the acceleration as a sub-panel with events ####
    ax = fig.add_subplot(gs[0, 4])

    # Plot acceleration if it exists
    if profile.acceleration is not Sensor.UNAVAILABLE:
        plot_events(ax, profile.events, plot_type='normal')
        for c in profile.acceleration_names:
            alpha = 1.0
            if c != profile.motion_detect_name:
                alpha = 0.3
            style = SensorStyle.from_column(c)
            ax.plot(profile.time, profile.raw[c], color=style.color,
                    label=style.label, alpha=alpha)

        ax.set_ylabel("Acceleration [g's]")
        ax.set_xlabel('Time [s]')
        ax.set_title('Accelerometer')
        ax.legend(fontsize='xx-small')
    else:
        ax.annotate('No Accelerometer \nData Available', (0.2, 0.5), alpha=0.5)
    #### plot the depth as a sub-panel with events ####
    ax = fig.add_subplot(gs[1, 4])
    ax.set_title('Depth')
    plot_events(ax, profile.events, plot_type='normal')

    # for label, color, col in label_color_column:
    if profile.motion_detect_name is Sensor.UNAVAILABLE:
        style = SensorStyle.from_column('barometer')
    else:
        acc_style = SensorStyle.ACCELERATION
        baro_style = SensorStyle.CONSTRAINED_BAROMETER
        raw = SensorStyle.RAW_BARO
        style = SensorStyle.FUSED
        ax.plot(profile.barometer.raw.index, profile.barometer.raw.values, color=raw.color, label=raw.label, alpha=0.5)
        ax.plot(profile.barometer.depth.index, profile.barometer.depth.values, color=baro_style.color, label=baro_style.label, alpha=0.5)
        ax.plot(profile.accelerometer.depth, color=acc_style.color, label=acc_style.label, alpha=0.5)

    ax.plot(profile.time, profile.depth, color=style.color, label=style.label)

    ax.legend(loc='upper right', fontsize='xx-small')
    # limits for depth
    buffer = 0.3
    n_samples = len(profile.raw)
    lim_idx1 = get_index_from_ratio(profile.start.index, 1- buffer, n_samples)
    lim_idx2 = get_index_from_ratio(profile.stop.index, 1 + buffer, n_samples)
    ts1, ts2 = profile.time[lim_idx1], profile.time[lim_idx2]
    depth1 = profile.depth.iloc[lim_idx1] * (1 - buffer)
    depth2 =  profile.depth.iloc[lim_idx2].min() * (1 + buffer)
    if abs(depth1) < 5:
        depth1 = 5

    # ax.set_xlim(ts1, ts2)
    # ax.set_ylim(*sorted([depth1, depth2]))
    ax.grid(True, axis='y', which='both', alpha=0.5)

    # Make the figure full screen
    manager = plt.get_current_fig_manager()
    if 'Linux' in platform.platform():
        manager.full_screen_toggle()

    # Manage plot time
    if timed_plot == 0:
        pass
    elif timed_plot is not None:
        timer = fig.canvas.new_timer(interval=timed_plot*1000)
        timer.add_callback(plt.close)
        timer.start()
        plt.show(block=True)

    else:
        plt.show()


def plot_hi_res_cli():
    files = sys.argv[1:]

    for f in files:
        try:
            plot_hi_res(fname=f)
        except Exception as e:
            print(f'Encountered error while processing {f}')
            traceback.print_exc(file=sys.stdout)
            time.sleep(1)
            plt.close()


def main():
    parser = argparse.ArgumentParser(
        description='Plot various versions of probe data.')
    parser.add_argument('file', help='path to measurement', nargs='+')

    parser.add_argument('--smooth''-sm', dest='smooth',
                        help='Provide a integer to describing number of windows to smooth over')

    parser.add_argument('--avg', '-a', dest='average', action='store_true',
                        help='Average all three sensors together')
    parser.add_argument('--compare', '-c', dest='compare', action='store_true',
                        help='Plots before and after sensors')
    parser.add_argument('--sensor', '-sn', dest='sensor', type=int,
                        help='plots only a specific sensor, must be between 1-4')

    args = parser.parse_args()

    # Provide a opportunity to look at lots
    filenames = []
    if args.file is not None and not isinstance(args.file, list):
        if os.path.isdir(args.file):
            filenames = os.listdir(args.file)
        elif os.path.isfile(args.file):
            filenames = [args.file]

    elif isinstance(args.file, list):
        filenames = args.file
    else:
        print("Please provide a directory or filename")
        sys.exit()

    for f in filenames:
        try:
            post_processed = False
            print('\n' + os.path.basename(f))
            print('=' * len(f))
            # Open the file and set the index to depth
            df_o, meta = read_csv(f)
            df = df_o.copy()

            if args.smooth is not None:
                df = df.rolling(window=int(args.smooth)).mean()
                post_processed = True

            print("Pre-processed profile:")
            print("\tNum of samples: {0:,}".format(len(df_o.index)))
            print("\tDepth achieved: {0:.1f}".format(min(df_o.index)[-1]))
            print("\tResolution: {0:.1f} pts/cm".format(
                abs(len(df_o.index) / min(df_o.index)[1] - max(df_o.index)[1])))

            if post_processed:
                print("\nPost-processed profile:")
                print("\tNum of samples: {0:,}".format(len(df.index)))
                print("\tDepth achieved: {0:.1f}".format(min(df.index)))
                print(
                    "\tResolution: {0:.1f} pts/cm".format(abs(len(df.index) / min(df.index) - max(df.index))))

            # Plot
            data = {}
            if args.average:
                data['Average'] = df[['Sensor 1',
                                      'Sensor 2', 'Sensor 3']].mean(axis=1)

            elif args.sensor is not None:
                if args.sensor in [1, 2, 3, 4]:
                    col = "SENSOR {0}".format(args.sensor)
                    data[col] = df[[col]]
            else:
                for i, c in enumerate(df.columns):
                    data[c] = df[c].copy()

            if args.compare:
                if args.average:
                    data['Orig. Average'] = df_o.mean(axis=1)
                else:
                    for i in range(1, 5):
                        col = 'SENSOR %s' % i
                        new_col = 'Orig. %s' % col
                        data[new_col] = df_o[col]

            # Parse the datetime
            for k, d in data.items():
                plt.plot(d, d.index, label=k)

            plt.title(os.path.basename(f))
            plt.legend()
            plt.xlabel('Reflectance')
            plt.ylabel('Depth from Surface [cm]')
            plt.xlim([0, 4100])
            plt.show()

        except Exception as e:
            print(e)


if __name__ == '__main__':
    plot_hi_res_cli()



from radicl.high_resolution import build_high_resolution_data
import pytest
from radicl.ui_tools import get_logger
import numpy as np
from . import MOCKCLI

class TestBuildingHighResolution:
    """
    This function is quite critical, so we test it in great detail here
    """
    @pytest.fixture(scope='class')
    def df(self):
        cli = MOCKCLI()
        log = get_logger('test_high_res')
        rawsensor = cli.grab_data('rawsensor')
        baro = cli.grab_data('filtereddepth')
        acc = cli.grab_data('rawacceleration')

        hi_res_df = build_high_resolution_data(rawsensor, baro, acc, log)
        return hi_res_df

    def test_no_nans(self, df):
        """ Ensure we have interpolated all nans"""
        assert ~np.any(np.isnan(df))

    @pytest.mark.parametrize('column, index, expected', [
        ('depth', 0, 0),
        ('depth', 19, -100),
        ('Y-Axis', 19, 2),
        ('Sensor1', 0, 1000),
        ('Sensor1', 19, 4000),
    ])
    def test_specific_value(self, df, column, index, expected):
        assert df[column].iloc[index] == expected




import time

import pytest
from time import sleep

from . import not_connected
import numpy as np
from radicl.probe import RAD_Probe
from radicl.info import ProbeState

from study_lyte.adjustments import merge_on_to_time

@pytest.mark.skipif(not_connected, reason='probe not connected')
class TestProbeMeasurementData:
    @pytest.fixture(scope='class')
    def meas_probe(self):
        prb = RAD_Probe(debug=True)
        prb.connect()
        prb.setSetting(setting_name='samplingrate', value=16000)
        prb.startMeasurement()
        time.sleep(1)
        prb.stopMeasurement()
        prb.wait_for_state(ProbeState.DATA_STAGED, retry=1000, delay=0.3)

        yield prb
        prb.resetMeasurement()

    @pytest.mark.skipif(not_connected, reason='probe not connected')
    def test_merging(self, meas_probe):
        baro = meas_probe.readFilteredDepthData()
        acc = meas_probe.readRawAccelerationData()
        final = merge_on_to_time([acc, baro], final_time=acc.index)

        # Check timing of adjusted dataset
        assert final['filtereddepth'].idxmax() == pytest.approx(baro['filtereddepth'].idxmax(), 1e-2)
        assert final['filtereddepth'].idxmin() == pytest.approx(baro['filtereddepth'].idxmin(), 1e-2)


@pytest.mark.skipif(not_connected, reason='probe not connected')
@pytest.mark.parametrize("setting_name, expected_type", [
    ('accrange', int)
])
def test_get_setting(real_probe, setting_name, expected_type):
    """
    Functionality test ensuring a setting runs
    """
    a = real_probe.getSetting(setting_name=setting_name)
    assert type(a) is expected_type

@pytest.mark.skipif(not_connected, reason='probe not connected')
@pytest.mark.parametrize("setting_name, value", [
    ('accrange', 2),
    ('samplingrate', 5000),
    ('zpfo', 80)
])
def test_set_setting(real_probe, setting_name, value):
    """
    Test setting a parameter in the probes settings
    """
    a = real_probe.setSetting(setting_name=setting_name, value=value)
    sleep(0.1)
    a = real_probe.getSetting(setting_name=setting_name)
    assert (a == value)


@pytest.mark.skipif(not_connected, reason='probe not connected')
@pytest.mark.parametrize('scale', [2, 16])
def test_rawacceleration_scaling(real_probe, scale):
    """
    If a real probe is connected, the magnitude of the sensor should
    always be very close to 1 unless the scaling is off.
    """
    a = real_probe.setSetting(setting_name='accrange', value=scale)
    time.sleep(0.1)
    real_probe.startMeasurement()
    time.sleep(1)
    real_probe.stopMeasurement()

    data = real_probe.readRawAccelerationData()

    data['magnitude'] = np.sqrt(np.power(data['X-Axis'], 2) +
                                np.power(data['Y-Axis'], 2) +
                                np.power(data['Z-Axis'], 2))
    result = np.mean(data['magnitude'])
    real_probe.resetMeasurement()
    time.sleep(0.2)
    assert pytest.approx(result, abs=5e-2) == 1

@pytest.mark.skipif(not_connected, reason='probe not connected')
def test_initial_state_set(real_probe):
    real_probe.getProbeMeasState()
    assert real_probe.state != ProbeState.NOT_SET

@pytest.mark.skipif(not_connected, reason='probe not connected')
def test_meas_temp(real_probe):
    temp = real_probe.readMeasurementTemperature()
    assert type(temp) == int

@pytest.mark.skipif(not_connected, reason='probe not connected')
def test_probe_serial(real_probe):
    result = real_probe.serial_number
    # Ensure the probe returns something real
    assert result is not None
    # Its a string
    assert type(result) == str
    # Its number and uppers
    assert all([c.isupper() or c.isnumeric() for c in result])

@pytest.mark.skipif(not_connected, reason='probe not connected')
def test_probe_header(real_probe):
    result = real_probe.getProbeHeader()
    assert result is not None
    assert type(result) == dict
    assert all([v is not None for k,v in result.items()])




import time
from pynmeagps import NMEAReader

from .com import get_serial_cnx
from .ui_tools import get_logger


class USBGPS:
    def __init__(self, debug=False):
        self.log = get_logger(__name__, debug=debug)
        try:
            self.cnx = get_serial_cnx('blox')
        except Exception as e:
            self.cnx = None
            self.log.error('Unable to open GPS port.')
            self.log.error(e)

        if self.cnx is None:
            self.log.warning('No GPS found. No location data will be recorded.')
        else:
            self.log.info(f'GPS found ({self.cnx.port})!')

    def get_fix(self, max_attempts=100):
        """
        Attempts to get a location fix given a serial port
        connection to a usb gps.

        Args:
            max_attempts: Number of attempts to get lat long

        Returns:
            location: tuple of latitude and longitude
        """

        location = None
        if self.cnx is not None:
            self.log.info('Attempting to get a fix on location...')

            gps = NMEAReader(self.cnx)
            for i in range(max_attempts):
                rx, msg = gps.read()
                if msg.msgID in ['GGA', 'GLL', 'RMC']:
                    info = msg.lat, msg.lon
                    if all(info):
                        location = [float(p) for p in info]
                        break
                time.sleep(0.1)

            if location is None:
                self.log.warning('Unable to get a fix on GPS! No location data will be recorded!')
            else:
                self.log.info(f'GPS fix acquired, {location[0]:0.6f} {location[1]:0.6f}')

        return location

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.cnx is not None:
            self.cnx.close()



import time
from radicl.gps import USBGPS
import pytest
from . import gps_not_connected


@pytest.mark.skipif(gps_not_connected, reason='gps not connected')
class TestUSBGPS:
    @pytest.fixture(scope='class')
    def gps_dev(self):
        return USBGPS()

    def test_get_fix(self, gps_dev):
        location = gps_dev.get_fix(max_attempts=30)

        assert len(location) == 2

    def test_get_fix_not_identical(self, gps_dev):
        """ Even a static GPS should be slightly different."""
        location = gps_dev.get_fix(max_attempts=30)
        time.sleep(0.5)
        location2 = gps_dev.get_fix(max_attempts=30)

        assert location != location2



# coding: utf-8

import datetime
import inspect
import sys
import time
from os.path import join, abspath, dirname, expanduser, isdir

import pandas as pd
from matplotlib import pyplot as plt
from termcolor import colored
from study_lyte.io import write_csv

from .probe import RAD_Probe
from .calibrate import get_avg_sensor
from .ui_tools import (Messages, get_logger, parse_func_list, parse_help,
                       print_helpme)
from .info import ProbeState, CLIState


out = Messages()

DEV_MODE = True

def dataframe_this(data, name=None):
    """
    Takes the data returned from a function and converts it into a DataFrame
    There appears to be two scenarios.
    Args:
        data: list or dictionary of data to dataframe
        name: Name to use in the event the data is a list
    Returns:
        df: pd.Dataframe of the provided data
    """
    t = type(data)
    if t == dict:
        df = pd.DataFrame.from_dict(data)

    elif t == list:
        if isinstance(data[0], tuple):
            data = [d[0] for d in data]

        df = pd.DataFrame(data, columns=[name])
    elif t == pd.DataFrame:
        df = data

    return df


def is_numbered(filename):
    """
    Checks if the filename has been numbered. Denoted by any file ending in
    somefname_<#>.csv Returns true or false.
    Args:
        filename: Path with potential some appended number scheme denoted by _
    Returns:
        bool: True if a separator and all numbers are found towards the end
    """
    info = filename.split('.')[0]
    sep = False
    numbered = False

    if '_' in info:
        numbers = info.split('_')[-1]
        sep = True
        numbered = all([c.isnumeric() for c in numbers])
    return sep and numbered


def add_ext(filename):
    """
    Check to see if the user provided the .csv ext in the filename
    and add it
    """
    if filename[-4:] != '.csv':
        f = filename.split('.')
        # Did the user try to add an ext
        if len(f) == 2:
            filename = f[0] + '.csv'
        else:
            filename += '.csv'

    return filename


def get_default_filename(output_dir='./'):
    """
    Creates a datetime path for writing to

    Returns:
        fname: csv path named by the datetime
    """

    t = datetime.datetime.now()
    fstr = "{0}-{1:02d}-{2:02d}--{3:02d}{4:02d}{5:02d}.csv"
    fname = fstr.format(t.year, t.month, t.day, t.hour, t.minute, t.second)
    return join(output_dir, fname)


def increment_fnumber(filename):
    """
    Check for a file numbering. Increment if there is one. Otherwise add one
    """
    no_ext = filename.split('.')[0]

    if is_numbered(filename):
        info = no_ext.split('_')
        base_file = '_'.join(info[0:-1])
        count = int(info[-1])
        fcount = count + 1

    else:
        base_file = no_ext
        fcount = 1

    filename = f"{base_file}_{fcount}.csv"

    return filename


class RADICL(object):
    """
    This is the main interface for end users to interact with the API and the
    probe.

    Attributes:
        probe: radicl
    """
    defaults = {'debug': False}

    def __init__(self, **kwargs):

        self.data = None
        self.current_setting_value = None

        self.state = CLIState.HOME
        for k, v in self.defaults.items():
            if k not in kwargs.keys():
                kwargs[k] = v

        self.log = get_logger(__name__, debug=kwargs['debug'])

        self.tasks = ['daq', 'settings', 'update']
        self.task_help = {'daq': "Data acquisition using the probe.",
                          'settings': 'Interface for modifying the behavior of the probe',
                          'update': 'Firmware update dialog for the probe.'}

        self.probe = RAD_Probe(debug=kwargs['debug'])
        self.probe.connect()

        self.running = True

        self.settings = dir(self.probe)

        # Get all the functions available in api and probe
        probe_funcs = inspect.getmembers(
            self.probe, predicate=inspect.ismethod)

        self.options = dict()

        # Assign all data functions with keywords to auto gather data packages
        self.options['data'] = parse_func_list(probe_funcs,
                                               ['read', 'Data'],
                                               ignore_keywords=['_by_segment', 'correlation', 'integrity'])

        # Grab the settings from the probe
        self.options['settings'] = self.probe.settings
        self.options['getters'] = self.probe.getters

        # Assign a single help statement for the help option
        self.options['settings']["help"] = self.print_settings

        # User runtime preferences
        self.output_preference = None
        self.filename = None
        self.daq = None

        # Gather help dialogs
        self.help_dialog = {}

        for s in self.options.keys():

            # We haven't added the category yet
            if s not in self.help_dialog:
                self.help_dialog[s] = {}

            # Add in all the functions
            for k, fn in self.options[s].items():
                self.help_dialog[s][k] = parse_help(fn.__doc__)

    def run(self):
        """
        Main running loop for the command line Interface
        """
        # Create the probe command with an arbitrary command to have access to
        while self.running:

            self.probe.getProbeMeasState()

            if self.state == CLIState.HOME:
                # If probe is not ready.
                if not ProbeState.ready(self.probe.state):
                    self.probe.resetMeasurement()

                else:
                    # Request type of operation
                    self.task = self.ask_user("What do you want to do with the"
                                              " probe?", self.tasks,
                                              helpme=self.task_help)
                    self.state = CLIState.from_choice(self.task)

            elif self.state != CLIState.HOME:
                self.tasking()

            else:
                pass

    def tasking(self):
        """
        Handles tasks from the user, specifically any member function who
        contains the keyword task_ in it. This means currently we are looking
        for a daq, settings, upgrade task function.
        """
        # Auto look for task_functions
        task_at = "task_%s" % self.task
        fn = getattr(self, task_at)

        # Call task_function
        self.log.debug("Using task function {0}".format(fn.__name__))
        fn()

    def calibrate(self):
        """
        Walks the user through calibrating
        """
        answer = self.ask_user(
            'Do you want to manually input or measure calibration values?',
            answer_lst=[
                'manual',
                'measure'])

        for i in range(1, 3):

            out.msg("For Sensor {}:".format(i))
            if answer == 'manual':
                hi = int(input('\tEnter high ADC value:\n\t'))
                lo = int(input('\n\tEnter low ADC value:\n\t'))

            elif answer == 'measure':
                input("\tApply the high value to the sensor and press enter:")
                hi = get_avg_sensor(self.probe, sensor='Sensor{}'.format(i))

                input("\n\tApply the low value to the sensor and press enter:")
                lo = get_avg_sensor(self.probe, sensor='Sensor{}'.format(i))

            else:
                raise ValueError("Invalid request for calibration.")

            self.log.info(
                "Setting calibration data to : low = {} hi = {}".format(
                    lo, hi))

            if i == 1:
                self.log.debug("Inverting for hardness...")
                low = hi
                hi = lo
                lo = low
            self.log.debug(
                "Setting calibration data to : low = {} hi = {}".format(
                    lo, hi))
            self.probe.api.MeasSetCalibData(i, lo, hi)

            values = self.probe.getSetting(setting_name='calibdata', sensor=i)
            self.log.debug("Calibdata now set to = {}".format(
                ", ".join([str(v) for v in values])))

    def take_a_reading(self):
        """
        Walks a user through the measurement taking process through pressing a
        key to start and stop measurements on the probe
        """

        input("Press any key to begin a measurement.\n")
        response = self.probe.startMeasurement()

        input("Press any key to stop the measurement.\n")
        response = self.probe.stopMeasurement()

    def listen_for_a_reading(self):
        """
        Simple CLI function to take a measurement by listening for a button
        triggering the start and stop on the probe.
        """

        out.msg(">> Press the probe button to start the measurement:")
        self.probe.wait_for_state(ProbeState.MEASURING, retry=1000, delay=0.3)
        out.respond("Measurement Started...")

        out.msg(">> Press the probe button to end the measurement:")
        self.probe.wait_for_state(ProbeState.DATA_STAGED, retry=1000, delay=0.3)
        out.respond("Measurement ended...")

    def grab_data(self, data_request, retries=3):
        """
        Grabs data from the probe and puts it into a dataframe

        Args:
            data_request: String name of the data requesting from probe, must be a key in the self.options
            retries: Number of times to attempt to retrieve data before throwing an error
        Returns:
            data: Dataframe of the data requested
        """
        fn = self.options['data'][data_request]

        self.log.info('Downloading {} data from probe...'.format(data_request))

        attempts = 0
        success = False

        while attempts < retries and not success:
            self.log.debug(
                "Requesting data using function {0}".format(
                    fn.__name__))

            # Enable debugging w/o try and except
            if DEV_MODE:
                df = fn()
                success = True
            else:
                try:
                    df = fn()
                    success = True

                except Exception as e:
                    self.log.warning(
                        'Failed to retrieve {} data, retrying...'.format(data_request))
                    self.log.debug(
                        "Failed {} attempt #{}".format(
                            data_request, attempts))
                    self.log.error(e)
                    success = False
                    attempts += 1

        if not success:
            m = ("Unable to retrieve {} data after {} attempts"
                 "".format(data_request, attempts))
            self.log.error(m)
            df = None
        return df

    def task_daq(self):
        """
        The state machine for request to do daq
        """
        # Data Selection
        if self.state == CLIState.DAQ_HOME:
            self.probe.getProbeMeasState()
            self.log.info("Checking to see if probe is ready...")
            self.log.debug(f"Probe State = {self.probe.state}")

            # Make sure probe is ready
            if ProbeState.ready(self.probe.state):
                self.daq = self.ask_user("What data do you want?",
                                         list(self.options['data'].keys()),
                                         helpme=self.help_dialog['data'])
                if self.state != CLIState.HOME:
                    self.state = CLIState.DAQ_CHOICE

        # Method for outputting data
        elif self.state == CLIState.DAQ_CHOICE:
            self.output_preference = self.ask_user("How do you want to output"
                                                   " {0} data?".format(
                self.daq),
                ['plot', 'write', 'both'],
                default_answer=self.output_preference)
            if self.state != CLIState.HOME:
                self.state = CLIState.DAQ_MEASUREMENT

        # Take Measurements
        elif self.state == CLIState.DAQ_MEASUREMENT:
            self.take_a_reading()
            self.data = self.grab_data(self.daq)

            if self.data is None:
                if self.state != CLIState.HOME:
                    self.state = CLIState.DAQ_MEASUREMENT
                self.log.error('')
                self.log.error('Retrieving probe data failed, try again.')
            else:
                if self.state != CLIState.HOME:
                    self.state = CLIState.DAQ_OUTPUT

            response = self.probe.resetMeasurement()

        # Data output and options
        elif self.state == CLIState.DAQ_OUTPUT:
            if self.output_preference in ['write', 'both']:
                valid = False

                #  Wait for real path
                while not valid:
                    fname = get_default_filename()
                    out.msg("Enter in a filepath for the data to be saved:")
                    filename = input("\nPress enter to use default:"
                                     "\n(Default: {0})".format(fname))

                    # Assign a default path
                    if filename == '':
                        filename = fname

                    # Make it absolute
                    filename = abspath(filename)
                    real_path = isdir(dirname(filename))

                    # Double-check a real path was given
                    if not real_path:
                        out.warn("Path provided does not exist.")
                        valid = False
                    else:
                        self.filename = filename
                        valid = True
                        self.log.info(
                            "Saving Data to :\n{0}".format(
                                self.filename))

                        if not self.data.empty:
                            self.write_probe_data(self.data, filename=filename)
                            self.state = CLIState.DAQ_FINISHED

            if self.output_preference in ['plot', 'both']:
                self.data.plot()
                plt.show()
                self.state = CLIState.DAQ_FINISHED

        elif self.state == CLIState.DAQ_FINISHED:
            response = self.ask_user("Take another measurement?")
            if response in ['y', 'yes', True]:
                self.state = CLIState.DAQ_MEASUREMENT

            # Go to the Home menu
            else:
                self.state = CLIState.HOME

    def task_settings(self):
        """
        Routine for handling users requests for modifying probe settings settings
        """

        # Setting Selection
        if self.state == CLIState.SETTINGS_HOME:
            self.probe.getProbeMeasState()
            self.log.info("Checking to see if probe is ready...")
            self.log.debug(f"Probe State = {self.probe.state}")

            # Make sure probe is ready
            if ProbeState.ready(self.probe.state):
                self.setting_request = \
                    self.ask_user("What setting do you want to adjust?",
                                  sorted(list(self.probe.settings.keys())),
                                  helpme=self.help_dialog['settings'])
            if self.state != CLIState.HOME:
                self.state = CLIState.SETTINGS_CHOICE

        # Get current setting
        elif self.state == CLIState.SETTINGS_CHOICE:

            if self.setting_request == 'show':
                self.options['settings'][self.setting_request]()
                self.state = CLIState.SETTINGS_HOME

            # Retrieve calibration values
            elif self.setting_request == 'calibdata':
                # TODO for calibration update this needs changing
                for i in range(1, 3):
                    self.current_setting_value = []
                    self.current_setting_value.append(
                        self.probe.getSetting(setting_name=self.setting_request,
                                              sensor=i))
                    values = ", ".join([str(v)
                                        for v in self.current_setting_value])
                    msg = ("Currently {0}[{2}] = {1}\nEnter value to change probe {0}\n"
                           "".format(self.setting_request,
                                     values, i))
                    self.log.info(msg)

                if self.state != CLIState.HOME:
                    self.state = CLIState.MODIFY_SETTING

            else:
                # Retrieve the current setting
                self.current_setting_value = \
                    self.probe.getSetting(setting_name=self.setting_request)
                if self.state != CLIState.HOME:
                    self.state = CLIState.MODIFY_SETTING


        # Modify setting
        elif self.state == CLIState.MODIFY_SETTING:
            if self.setting_request == 'calibdata':
                self.calibrate()
                if self.state != CLIState.HOME:
                    self.state = CLIState.SETTINGS_HOME

            else:
                msg = (f"Currently {self.setting_request} = {self.current_setting_value}\n"
                       f"Enter value to change probe {self.setting_request} [{self.current_setting_value}]\n")
                self.log.info(msg)

                valid = False

                # All entries have to be numeric. Ensure this is the case.
                while not valid:
                    self.new_value = input(msg)
                    if self.new_value == '':
                        self.new_value = self.current_setting_value
                    try:
                        self.new_value = int(self.new_value)
                        valid = True
                    except Exception as e:
                        out.error(e)
                        out.error("Value must be numeric!")

                # Call the function to change the setting
                self.probe.setSetting(setting_name=self.setting_request,
                                      value=self.new_value)
                time.sleep(0.2)

                # Confirm the value was changed
                test_value = self.probe.getSetting(
                    setting_name=self.setting_request)

                if test_value == self.new_value:
                    self.log.info("{0} was changed from {1} to {2}!\n"
                                  "".format(self.setting_request,
                                            self.current_setting_value,
                                            self.new_value))
                time.sleep(1)

                # Go back to settings menu
                if self.state != CLIState.HOME:
                    self.state = CLIState.SETTINGS_HOME

    def print_settings(self):
        """
        print all the current settings

        helpme - Prints out all the settings for the probe

        """

        msg = "\n===== Current Probe Settings =====\n"

        for s, fn in self.probe.getters.items():
            if s.lower() not in ['calibdata', 'numsegments']:
                value = self.probe.getSetting(setting_name=s)
                msg += "{} = {}\n".format(s, value)
        out.msg(msg)
        time.sleep(2)

    def write_probe_data(self, df, filename='', extra_meta={}):
        """
        Writes out a dataframe with a probe header to csv
        Args:
            df: pandas dataframe containing data
            filename: valid path to output to, if empty uses datetime
            extra_meta: Dictionary of extra notes to add to the file header
        """

        # Receive a default request
        if filename == '':
            filename = get_default_filename()

        else:
            filename = expanduser(filename)

        out.msg("Saving Data to :\n{0}".format(filename))

        if not df.empty:
            # Write the header so we know things about this
            meta = self.probe.getProbeHeader()
            meta.update(extra_meta)
            write_csv(df, meta, filename)
        return filename

    def ask_user(self, question_str, answer_lst=None, helpme=None,
                 default_answer=None):
        """
        Function is used to wait for an appropriate response from user and handle unknown answers

        Args:
            question_str - String of the question
            answer_lst - list of acceptable answers
            helpme - help strings in a dictionary to elaborate on acceptable answers
            next_state - Advance the state of the cli state machine
            default_answer - if the user selects nothing and hits enter the answer will default to this

        Returns:
            response - string of the valid response from the user
        """

        # Ask user for a yes no question
        if answer_lst is None:
            answer_lst = ['y', 'n']
            question_boolean = True
        else:
            question_boolean = False

        # Keep question to one line if small number of options
        lower_lst = [i.lower() for i in sorted(answer_lst)]

        # Add in option to return to main menu
        if self.state != CLIState.HOME:
            lower_lst.append('home')

        # Help documentation
        if helpme is not None:
            lower_lst.append("help")

        # Exit the program
        lower_lst.append('exit')

        # If the options list if small keep it on the same line
        if len(lower_lst) <= 5:
            print_able = " ("
            print_able += ", ".join(lower_lst)
            print_able += ")"

        # If the list of choices is long create a column of choices
        else:
            print_able = '\n\n{0}'.format(colored('[OPTIONS]', 'magenta',
                                                  attrs=['bold']))
            for s in lower_lst:
                print_able += '\n  {0}'.format(s)

        question_str += print_able

        # Provide prompt for default options
        if default_answer is not None:
            if default_answer not in answer_lst:
                raise ValueError("Default answer must be a member of the "
                                 "acceptable answers list in ask_user function")

            question_str += '(Default: {0})'.format(default_answer)

        out.msg(question_str)

        acceptable_answer = False

        while not acceptable_answer:
            user_answer = input('\n')
            response = user_answer.lower().strip()

            # User requests help with options
            if "help" in response:
                print_helpme(response, helpme)
                acceptable_answer = False
                out.msg(question_str)

            # Request to leave the program
            elif response == 'exit':
                self.log.info("Exiting the RAD CLI.")
                self.running = False
                sys.exit()

            elif response == 'home':
                self.log.info("Returning to the main menu.")
                self.running = True
                acceptable_answer = True
                self.state = CLIState.HOME

            # Default Picked
            elif response == '':
                response = default_answer

            # User provided a potential acceptable answer
            elif response in lower_lst:
                acceptable_answer = True
                response = lower_lst[lower_lst.index(response)]
                if question_boolean:
                    if response == 'y' or response == 'yes':
                        response = True
                    else:
                        response = False

            else:
                out.error('Only acceptable answers are:')
                out.msg(print_able)
                acceptable_answer = False

        return response



from radicl.interface import RADICL, dataframe_this, is_numbered, get_default_filename, add_ext, increment_fnumber
import pytest
import pandas as pd
from unittest.mock import patch
import datetime
from . import MockProbe


@pytest.mark.parametrize('data, name, expected', [
    ({"sensor": [1, 2, 3]}, None, pd.DataFrame.from_dict({"sensor": [1, 2, 3]})),
    ([1, 2, 3], 'sensor', pd.DataFrame.from_dict({"sensor": [1, 2, 3]}))
])
def test_dataframe_this(data, name, expected):
    df = dataframe_this(data, name=name)
    pd.testing.assert_frame_equal(df, expected)


@pytest.mark.parametrize('filename, expected', [
    ('test_1000.csv', True),
    ('test_1000_numbers.csv', False),
    ('test.csv', False),
])
def test_is_numbered(filename, expected):
    result = is_numbered(filename)
    assert result == expected


@pytest.mark.parametrize('filename, expected', [
    ('test.csv', 'test.csv'),
    ('test', 'test.csv'),
])
def test_add_ext(filename, expected):
    result = add_ext(filename)
    assert result == expected


def test_get_default_filename():
    dt = datetime.datetime(2023, 1, 1, 1, 2, 3)
    with patch('radicl.interface.datetime.datetime') as mock_date:
        mock_date.now.return_value = dt
        filename = get_default_filename()
    assert filename == './2023-01-01--010203.csv'


@pytest.mark.parametrize('filename, expected', [
    ('test.csv', 'test_1.csv'),
    ('test_test_10.csv', 'test_test_11.csv'),
])
def test_increment_fnumber(filename, expected):
    result = increment_fnumber(filename)
    assert result == expected



from enum import Enum


class ProbeState(Enum):
    """ States for the probe running during a measurement"""
    NOT_SET = 10
    IDLE = 0
    MEASURING = 1
    PROCESSING = 2
    DATA_STAGED = 3
    SENDING_DATA = 4
    RESET = 5
    UNKOWN_STATE = 11


    @classmethod
    def from_state(cls, state):
        final = cls.UNKOWN_STATE
        for e in cls:
            if e.value == state:
                final = e
                break
        return final

    @classmethod
    def ready(cls, state):
        """Function to determine if the probe is in an idle state"""
        return state in [cls.IDLE, cls.RESET]

    def __ge__(self, other):
        return self.value >= other.value

    def __le__(self, other):
        return self.value <= other.value

    def __lt__(self, other):
        return self.value < other.value

    def __gt__(self, other):
        return self.value > other.value


class CLIState(Enum):
    """Enum for managing the cli state"""
    HOME = 0
    DAQ_HOME = 1
    DAQ_CHOICE = 2
    DAQ_MEASUREMENT = 3
    DAQ_OUTPUT = 4
    DAQ_FINISHED = 5
    SETTINGS_HOME = 6
    SETTINGS_CHOICE = 7
    MODIFY_SETTING = 8
    UNKNOWN = 99
    @classmethod
    def from_choice(cls, choice):
        """Return state based on user input"""
        result = cls.UNKNOWN
        if choice == 'daq':
            result = cls.DAQ_HOME
        elif choice == 'settings':
            result = cls.SETTINGS_HOME
        return result

class ProbeErrors(Enum):
    # Define by code and error string
    MEASUREMENT_NOT_RUNNING = 2049,  "The probe measurement/sensor is not running"
    GENERIC_ERROR =  2048, 'Generic Measurement error'
    PROBE_NOT_READY = 2050, "Measurement FSM not ready to start measurement"
    PROBE_ALREADY_STOPPED = 2051, " Measurement FSM already stopped (i.e. in ready/idle state)"
    PROBE_FAILED_TO_SET_PARAM = 2052, "Setting a parameter failed",
    INVALID_DATA_BUFFER = 2053, "Invalid buffer was addressed"
    FAILED_DATA_READ = 2054, "Reading data failed"
    UNKNOWN_ERROR = None, "Unrecognized error code"

    @property
    def error_code(self):
        return self.value[0]

    @property
    def error_string(self):
        return self.value[1]


    @classmethod
    def from_code(cls, code):
        final = cls.UNKNOWN_ERROR
        for e in cls:
            if e.error_code == code:
                final = e
                break
        return final


    @classmethod
    def to_dict(cls):
        final = {}
        for e in cls:
            final[e.error_code] = e.error_string
        return final


class AccelerometerRange(Enum):
    # From Data sheet for accelerometer in REV C
    RANGE_2G = 2, 0.06
    RANGE_4G = 4, 0.12
    RANGE_6G = 6, 0.18
    RANGE_8G = 8, 0.24
    RANGE_16G = 16, 0.73

    @property
    def sensing_range(self):
        return self.value[0]

    @property
    def value_scaling(self):
        return self.value[1]

    @classmethod
    def from_range(cls, sensing_range):
        final = None
        for e in cls:
            if e.sensing_range == sensing_range:
                final = e
                break
        return final


class SensorReadInfo(Enum):

    # Buffer Index, bytes per seg, number of bytes per value, expected number of values, does the data live in spi mem.
    RAWSENSOR = 0, 256, 2, 4, True, 'Raw Sensor', ["Sensor1", "Sensor2", "Sensor3", "Sensor4"], None, None, 16000
    # 3 axis acc, convert from mG to G
    ACCELEROMETER = 1, None, 2, 3, False, 'Acceleration', ["X-Axis", "Y-Axis", "Z-Axis"], '<h', 0.001, 100
    RAW_BAROMETER_PRESSURE = 2, 256, 3, 1, False, 'Raw Pressure', ['raw_pressure'], None, None, 75
    FILTERED_BAROMETER_DEPTH = 4, None, 4, 1, False, 'Filtered Barometer Depth', ['filtereddepth'], 'f', 0.01, 75


    @property
    def buffer_id(self):
        return self.value[0]

    @property
    def bytes_per_segment(self):
        return self.value[1]

    @property
    def nbytes_per_value(self):
        return self.value[2]

    @property
    def expected_values(self):
        return self.value[3]

    @property
    def uses_spi(self):
        return self.value[4]

    @property
    def readable_name(self):
        return self.value[5]

    @property
    def data_names(self):
        return self.value[6]

    @property
    def unpack_type(self):
        return self.value[7]

    @property
    def conversion_factor(self):
        return self.value[8]

    @property
    def max_sample_rate(self):
        return self.value[9]

    @property
    def bytes_per_sample(self):
        """ Number of bytes per sample"""
        return self.nbytes_per_value * self.expected_values

    @classmethod
    def from_data_request(cls, data_request):
        if data_request == 'filtered_depth':
            result = cls.FILTERED_BAROMETER_DEPTH
        elif data_request in ['calibratedsensor', 'rawsensor']:
            result = cls.RAWSENSOR
        elif data_request == 'rawpressure':
            result = cls.RAW_BAROMETER_PRESSURE
        elif data_request == 'rawacceleration':
            result = cls.ACCELEROMETER
        else:
            result = None
        return result


class Firmware:
    """
    Small firmware class for comparing firmwares
    """
    def __init__(self, firmware_str):
        if firmware_str is not None:
            self._info = firmware_str.split('.')
        else:
            self._info = (-1,-1,-1,-1)
        self.sub_versions = ['major_version', 'minor_version', 'patch_version', 'build_number']
        # Default to 0
        self.major_version = 0
        self.minor_version = 0
        self.patch_version = 0
        self.build_number = 0
        # Assign if the value was provided, assumed order ie. 1.3 == 1.3.0.0
        for i, version in enumerate(self.sub_versions):
            if i < len(self._info):
                setattr(self, version, int(self._info[i]))

    def __eq__(self, other):
        comparisons = [getattr(self, version) == getattr(other, version) for version in self.sub_versions]
        return all(comparisons)

    def __ge__(self, other):
        gt_comparisons = [getattr(self, version) > getattr(other, version) for version in self.sub_versions]
        eq_comparisons = [getattr(self, version) == getattr(other, version) for version in self.sub_versions]
        result = False
        for greater_than, equal in zip(gt_comparisons, eq_comparisons):
            if greater_than:
                result = True
                break
            elif not greater_than and not equal:
                result = False
                break
            elif equal:
                result = True

        return result

    def __gt__(self, other):
        result = False
        for  version in self.sub_versions:
            v = getattr(self, version)
            ov = getattr(other, version)
            if v > ov:
                result = True
                break

        return result

    def __repr__(self):
        return f"v{self.major_version}.{self.minor_version}.{self.patch_version}.{self.build_number}"


class PCA_Name(Enum):
    UNKNOWN = None
    PB1 = 1
    PB2 = 2
    PB3 = 3

    @classmethod
    def from_index(cls, idx):
        result = cls.UNKNOWN
        for e in cls:
            if e.idx == idx:
                result = e
                break
        return result

    @property
    def idx(self):
        return self.value



from radicl.info import AccelerometerRange, ProbeErrors, ProbeState, Firmware, PCA_Name, SensorReadInfo
import pytest


class TestAccelerometerRange:
    @pytest.mark.parametrize('sensing_range, expected', [
        (2, AccelerometerRange.RANGE_2G),
        (3, None)
    ])
    def test_from_range(self, sensing_range, expected):
        assert AccelerometerRange.from_range(sensing_range) == expected


class TestProbeErrors:
    @pytest.mark.parametrize('error_code, expected', [
        (2049, ProbeErrors.MEASUREMENT_NOT_RUNNING),
        (10, ProbeErrors.UNKNOWN_ERROR)
    ])
    def test_from_code(self, error_code, expected):
        assert ProbeErrors.from_code(error_code) == expected


class TestFirmware:

    @pytest.fixture(scope='function')
    def fw1(self, fw1_str):
        return Firmware(fw1_str)

    @pytest.fixture(scope='function')
    def fw2(self, fw2_str):
        return Firmware(fw2_str)

    @pytest.mark.parametrize('fw1_str, fw2_str, expected', [
        ('1.2.3.0', '1.2.3.0', True),
        # Check auto assume subsequent zero values
        ('1.0.0.0', '1', True)

    ])
    def test_firmware_eq(self, fw1, fw2, expected):
        comparison = (fw1 == fw2)
        assert comparison == expected

    @pytest.mark.parametrize('fw1_str, fw2_str, expected', [
        # Test greater at both ends of the scales
        ('2.0', '1.1', True),
        ('0.0.0.2', '0.0.0.1', True),
        # Test equal
        ('1.2.3', '1.2.3', True),
        # Test greater value in a different position
        ('0.2', '1.1', False),
    ])
    def test_firmware_ge(self, fw1, fw2, expected):
        comparison = (fw1 >= fw2)
        assert comparison == expected

    @pytest.mark.parametrize('fw1_str, fw2_str, expected', [
        ('1.2', '1.1', True),
        ('0.2', '0.2', False),
        ('0.0.0.2', '0.0.0.1', True),
        ('0.0.0.1', '0.0.0.2', False),

    ])
    def test_firmware_gt(self, fw1, fw2, expected):
        comparison = (fw1 > fw2)
        assert comparison == expected

    @pytest.mark.parametrize('fw1_str, expected', [
        ('1.2', 'v1.2.0.0')
    ])
    def test_repr(self, fw1, expected):
        assert str(fw1) == expected

    def test_null_fw(self):
        assert Firmware(None) == Firmware('-1.-1.-1.-1')


@pytest.mark.parametrize('idx, expected', [
    (1, PCA_Name.PB1),
    (3, PCA_Name.PB3),
    (100, PCA_Name.UNKNOWN),
    (None, PCA_Name.UNKNOWN)
])
def test_PCA_NAME(idx, expected):
    assert PCA_Name.from_index(idx) == expected


class TestProbeState():
    @pytest.mark.parametrize("state, expected", [
        (1, ProbeState.MEASURING),
        (100, ProbeState.UNKOWN_STATE),

    ])
    def test_from_state(self, state, expected: ProbeState):
        assert ProbeState.from_state(state) == expected

    @pytest.mark.parametrize("state1, state2, expected", [
        (ProbeState.IDLE, ProbeState.MEASURING, True),
        (ProbeState.IDLE, ProbeState.IDLE, True),
        (ProbeState.DATA_STAGED, ProbeState.IDLE, False),

    ])
    def test_le(self, state1, state2, expected):
        assert (state1 <= state2) == expected

    @pytest.mark.parametrize("state1, state2, expected", [
        (ProbeState.MEASURING, ProbeState.MEASURING, True),
        (ProbeState.SENDING_DATA, ProbeState.IDLE, True),
        (ProbeState.PROCESSING, ProbeState.DATA_STAGED, False),

    ])
    def test_ge(self, state1, state2, expected):
        assert (state1 >= state2) == expected

    @pytest.mark.parametrize("state, expected", [
        (ProbeState.MEASURING, False),
        (ProbeState.IDLE, True),
        (ProbeState.RESET, True),
    ])
    def test_ready(self, state, expected):
        assert ProbeState.ready(state) == expected

class TestReadSensorInfo:
    @pytest.mark.parametrize("sensor_info, att, expected", [
        (SensorReadInfo.RAWSENSOR, 'buffer_id', 0),
        # Bytes per segment
        (SensorReadInfo.ACCELEROMETER, 'bytes_per_segment', None),
        (SensorReadInfo.RAWSENSOR, 'bytes_per_segment', 256),
        # nbytes_per_value
        (SensorReadInfo.RAW_BAROMETER_PRESSURE, 'nbytes_per_value', 3),
        # expected values
        (SensorReadInfo.FILTERED_BAROMETER_DEPTH, 'expected_values', 1),
        (SensorReadInfo.ACCELEROMETER, 'expected_values', 3),
        # spi usage
        (SensorReadInfo.RAWSENSOR, 'uses_spi', True),
        (SensorReadInfo.ACCELEROMETER, 'uses_spi', False),
        # Readable name
        (SensorReadInfo.ACCELEROMETER, 'readable_name', 'Acceleration'),
        # data_names
        (SensorReadInfo.RAW_BAROMETER_PRESSURE, 'data_names', ['raw_pressure']),
        # unpack type
        (SensorReadInfo.ACCELEROMETER, 'unpack_type', '<h'),
        # conversion
        (SensorReadInfo.ACCELEROMETER, 'conversion_factor', 0.001),
        # Max sample
        (SensorReadInfo.RAWSENSOR, 'max_sample_rate', 16000),
    ])
    def test_properties(self, sensor_info:SensorReadInfo, att, expected):
        result = getattr(sensor_info, att)
        assert result == expected

    @pytest.mark.skip("Unsure why this wont evaluate")
    def test_bytes_per_sample(self):
        result = SensorReadInfo.bytes_per_sample
        assert result == 8

    @pytest.mark.parametrize("data_request, expected", [
        ('rawsensor', SensorReadInfo.RAWSENSOR),
        ('calibratedsensor', SensorReadInfo.RAWSENSOR),
        ('rawacceleration', SensorReadInfo.ACCELEROMETER),
        ('rawpressure', SensorReadInfo.RAW_BAROMETER_PRESSURE),
        ('garbage', None),
    ])
    def test_from_data_request(self, data_request, expected):
        result = SensorReadInfo.from_data_request(data_request)
        assert result == expected



# coding: utf-8

import binascii
import hashlib
import time
from enum import Enum
from .ui_tools import get_logger
from .info import Firmware
from .api import RAD_API


class FWState(Enum):
    """Mirror class of the FW update states"""
    IDLE = 0
    PREPARE = 1
    READY_TO_UPDATE = 2
    DOWNLOAD = 3
    DOWNLOAD_COMPLETE = 4
    VERIFICATION = 5
    DONE = 6
    ERROR = 7
    # If we cant id the state use this
    UNKNOWN = None

    @classmethod
    def from_int(cls, state_int):
        result = cls.UNKNOWN

        for e in cls:
            if e.value == state_int:
                result = e
                break
        return result


class FW_Update:

    # *********************
    # * PRIVATE FUNCTIONS *
    # *********************

    def __init__(self, api:RAD_API):
        self.f = None
        self.api = api
        self.file_crc = 0
        self._packet_size = None
        self.log = get_logger(__name__, debug=True)
        self.state = FWState.UNKNOWN

    @property
    def packet_size(self):
        """ Get the packet size based on the fw"""

        if self._packet_size is None:
            if self.api.fw_rev >= Firmware('1.45'):
                self._packet_size = 64
            else:
                self._packet_size = 16
        return self._packet_size

    def __getFileSize(self):
        """
        Returns the file size of the specified file
        """

        self.f.seek(0, 2)  # move the cursor to the end of the file
        size = self.f.tell()
        return size

    def __swapBytesInWord(self, fourBytes):
        """
        Helper function to swap a word (4 bytes)
        """

        return [fourBytes[3], fourBytes[2], fourBytes[1], fourBytes[0]]

    def __swapBytesInArray(self, byteArray):
        """
        Helper function to swap bytes in an array (LSByte to MSByte)
        """

        length = len(byteArray)
        swappedArray = bytearray(length)
        index = 0
        offset = 0
        four_count = 4
        for b in byteArray:
            swappedArray[offset + four_count - 1] = byteArray[index]
            index += 1
            four_count = four_count - 1
            if four_count == 0:
                offset = index
                four_count = 4
        return swappedArray

    def __LoadFileHeader(self):
        """
        Loads/reads the file header
        """

        self.f.seek(2048, 0)
        bytes = self.f.read(4)
        hex_string1 = "".join("%02x" %
                              b for b in self.__swapBytesInWord(bytes))
        bytes = self.f.read(4)
        hex_string2 = "".join("%02x" %
                              b for b in self.__swapBytesInWord(bytes))

        if hex_string1 == "dabbad00" and hex_string2 == "a5b6c7d8":
            bytes = self.f.read(4)
            hex_string = "".join("%02x" %
                                 b for b in self.__swapBytesInWord(bytes))

            self.app_header_version = int(hex_string, 16)
            bytes = self.f.read(4)
            hex_string = "".join("%02x" %
                                 b for b in self.__swapBytesInWord(bytes))

            self.app_header_length = int(hex_string, 16)
            bytes = self.f.read(4)
            hex_string = "".join("%02x" %
                                 b for b in self.__swapBytesInWord(bytes))

            self.app_header_entry = int(hex_string, 16)
            bytes = self.f.read(4)
            hex_string = "".join("%02x" %
                                 b for b in self.__swapBytesInWord(bytes))

            self.app_header_crc = int(hex_string, 16)
            return 1

        # Header magic words did not match.
        else:
            return 0

    def __DumpHeaderInfo(self):
        """
        Dumps the header info
        """

        self.log.info("*** Header Info ***"
                      "\nVersion\t=\t%d"
                      "\nApp length =\t%d"
                      "\nEntry\t=\t0x%0.8X"
                      "\nCRC\t=\t0x%0.8X" % (self.app_header_version,
                                             self.app_header_length,
                                             self.app_header_entry,
                                             self.app_header_crc))
        return 0

    def __CalculateAppCRC(self):
        # Set the file pointer to the beginning of the application section
        self.f.seek(self.app_header_entry, 0)
        # Read the entire application section from the file
        app_bytes = self.f.read(self.app_header_length)
        byte_string = "".join("%02x" %
                              b for b in self.swapBytesInWord(app_bytes))
        self.log.debug(
            "CRC non-swapped =\t0x%.08X, 0x%.08X" %
            (binascii.crc32(
                app_bytes,
                0x4C11DB7),
             (0x100000000 -
              binascii.crc32(
                  app_bytes,
                  0x4C11DB7))))
        self.log.debug(
            "CRC swapped =\t0x%.08X, 0x%.08X" %
            (binascii.crc32(
                self.__swapBytesInArray(app_bytes),
                0x4C11DB7),
             (0x100000000 -
              binascii.crc32(
                  self.__swapBytesInArray(app_bytes),
                  0x4C11DB7))))

        # return sum(app_bytes)
        return binascii.crc32(app_bytes)

    def __CalculateImageCRC32(self):
        """
        Calculates the CRC32 of the entire FW image file
        """

        file_len = self.__getFileSize()
        self.f.seek(0, 0)
        data = self.f.read(file_len)
        return binascii.crc32(data)

    def __CalculateChecksum(self, data):
        """
        Calculates the checmsum of the supplied data package
        """

        s = sum(data)
        return s % 256

    def __calculateNumDataPackets(self):
        """
        Calculates the number of data packets based on the file size
        """

        file_size = self.__getFileSize()
        num_packets = file_size / self.packet_size
        self.num_packets = num_packets
        return num_packets

    # ********************
    # * PUBLIC FUNCTIONS *
    # ********************

    def loadFile(self, file_name):
        """
        Loads and opens a FW update file
        This also verifies the file integrity and CRC
        Returns 1 if operation succeeded
        """

        try:
            self.f = open(file_name, "rb")
        except Exception as e:
            self.log.error(e)
            self.f = None
            return 0
        else:
            self.file_crc = self.__CalculateImageCRC32()
            self.log.debug("File CRC32 = 0x%08X" % self.file_crc)
            hash_md5 = hashlib.md5()
            self.f.seek(0, 0)
            for chunk in iter(lambda: self.f.read(4096), b""):
                hash_md5.update(chunk)
            self.log.debug("MD5 = %s" % hash_md5.hexdigest())

            # If we get here then the file is valid and can be used
            return 1

    def closeFile(self):
        """
        Closes the file
        """

        self.f.close()
        self.f = None

    def getState(self):
        """
        Returns the current state of the update FSM
        Returns the state if successful, None otherwise
        """
        ret = self.api.UpdateGetState()
        state = FWState.UNKNOWN

        if ret['status'] == 1:
            byte_arr = ret['data']
            state_int = int.from_bytes(byte_arr, byteorder='little')
            state = FWState.from_int(state_int)
            self.log.info(f"State = {state}({state_int})")
        else:
            if ret['errorCode'] is not None:
                self.log.error(
                    "FW_Update.getState returned error %d" %
                    ret['errorCode'])
            state = FWState.ERROR

        self.state = state

        return self.state

    def waitForStateChange(self, timeout):
        """
        Waits for an update FSM state change message
        Returns the new state if successful, None if timeout or incorrect
        response
        """
        state = FWState.UNKNOWN
        ret = self.api.UpdateWaitForStateChange(timeout)

        if ret['status'] == 1:
            byte_arr = ret['data']
            state_int = int.from_bytes(byte_arr, byteorder='little')
            state = FWState.from_int(state_int)

        # Retry in case of timeout
        else:
            self.log.debug("Timeout - Retrying!")
            state = self.getState()

        if ret['errorCode'] is not None:
            self.log.error("FW_Update.waitForStateChange returned error %d" %
                           ret['errorCode'])
            state = FWState.ERROR
        return state

    def closeFSM(self):
        """
        Closes the update FSM (i.e. reset)
        Returns 1 if successful, 0 otherwise
        """

        ret = self.api.UpdateClose()

        if ret['status'] == 1:
            return 1

        if ret['errorCode'] is not None:
            self.log.error(
                "FW_Update.closeFSM returned error %d" %
                ret['errorCode'])
        return 0

    def enterFSM(self):
        """
        Enter the FW update FSM (starts the process)
        Returns 1 if successful, 0 otherwise
        """
        ret = self.api.UpdateEnter()

        if ret['status'] == 1:
            self.log.info(
                "Enter bootloader mode - Waiting for Flash memory to be cleared. This could take a few seconds")
            state = self.waitForStateChange(10)

            if state is None:
                self.log.error("Timeout: No response")
                return 0

            else:
                if state == FWState.READY_TO_UPDATE:
                    self.log.info(
                        "Probe is in bootloader mode. Flash memory is ready!")
                    return 1

                self.log.error(f"Incorrect state ({state}")
                return 0

        if ret['errorCode'] is not None:
            self.log.error(
                "FW_Update.enterFSM returned error %d" %
                ret['errorCode'])
        self.log.error("Entering bootloader mode failed!!")
        return 0

    def sendNumPackets(self):
        """
        Sends the number of data packets and ensures we advance to the next step
        Returns 1 if successful, 0 otherwise
        """
        state = self.getState()

        if state == FWState.READY_TO_UPDATE:
            self.__calculateNumDataPackets()
            self.log.debug("Total number of packets = %d" % self.num_packets)
            ret = self.api.UpdateSetSize(int(self.num_packets),
                                         self.packet_size)
            if ret['status'] != 1:
                self.log.error("Unable to set number of packets")
                if ret['errorCode'] is not None:
                    self.log.error("FW_Update.sendNumPackets returned error %d" %
                                   ret['errorCode'])
                return 0

            time.sleep(1)
            state = self.getState()

            if state is None:
                self.log.error("Timeout - No response")
                return 0
            elif state != FWState.DOWNLOAD:
                self.log.error(f"Unexpected state ({state})")
                return 0
            else:
                self.log.info("Number of packages set and accepted")
                return 1
        else:
            if state is None:
                self.log.error("Incorrect or no response from device...")
            else:
                self.log.error(f"Incorrect state ({state})")
            return 0

    def download_packet(self, packet_id):
        """Attempts to download a single packet to the device for fw update"""
        data = self.f.read(self.packet_size)
        packet_id_arr = packet_id.to_bytes(4, byteorder='little')
        retry = True
        retry_count = 0
        error = True

        while retry:
            crc8 = self.__CalculateChecksum(data)
            if self.packet_size == 256:
                ret = self.api.UpdateDownload_Long(data,
                                                   crc8,
                                                   packet_id_arr[0])
            else:
                ret = self.api.UpdateDownload_Short(data,
                                                    crc8,
                                                    packet_id_arr[0])

            if ret['status'] != 1:
                # Packet was not accepted. Check if it was checksum
                # mismatch
                if ret['errorCode'] == 5121:
                    if retry_count > 3:
                        self.log.error(
                            "Max retries exceeded. Stopping.")
                        retry = False
                        error = True
                    else:
                        # Retry, by simply allowing to loop again
                        retry = True
                        retry_count = retry_count + 1
                        self.log.error(f"Checksum mismatch error. Retrying packet"
                                       f" {packet_id} (retry={retry_count})")

                # Error other than checksum mismatch.
                else:
                    self.log.error(f"Download error (Packet={packet_id})")
                    retry = False
                    error = True


            # Packet was accepted by device. Skip retry/ update
            # progress
            else:
                retry = False
                error = False

        return error

    def downloadFile(self):
        """
        Downloads the file in chunks
        Returns 1 if successful, 0 otherwise
        """
        state = self.getState()
        if state == FWState.DOWNLOAD:
            # Set the file pointer to the very beginning
            self.f.seek(0)
            packet_id = 0

            while packet_id < self.num_packets:
                error = self.download_packet(packet_id)
                if not error:
                    packet_id = packet_id + 1

            # If we get here then the entire download succeeded
            self.log.info("Download done - Waiting for state change")
            time.sleep(1)
            state = self.getState()
            if state is None:
                self.log.error("Timeout: No response")
                return 0
            elif state != FWState.DOWNLOAD_COMPLETE:
                self.log.error(f"Unexpected state ({state}")
                return 0
            # If we get here then the probe is in state 4 and we can move on
            return 1
        else:
            if state is None:
                self.log.error("Incorrect or no response from device...")
            else:
                self.log.error(f"Incorrect state ({state})")
            return 0

    def sendImageCRC32(self):
        """
        Sends the CRC32 of the entire FW image
        Returns 1 if successful, 0 otherwise
        """
        success = False
        # Check that we are in the right state
        state = self.getState()

        if state == FWState.DOWNLOAD_COMPLETE:
            ret = self.api.UpdateSetCRC(self.file_crc)

            if ret['status'] != 1:
                self.log.error("Error sending CRC32")

                if ret['errorCode'] is not None:
                    self.log.error("FW_Update.sendImageCRC32 returned error %d" %
                                   ret['errorCode'])
                success = False
            else:
                success = True
        else:
            self.log.warning(f"Probe not in the write state to set CRC, ({self.state})")
            success = False

        return success

    def verify(self):
        """Wait for the probe to respond after verification of the update image"""
        # Grab the state
        state = self.getState()

        # probe is still verifying
        if state == FWState.VERIFICATION:
            self.log.debug(
                "CRC32 set - Waiting for verification to complete")

        elif state == FWState.DONE:
            self.log.info(
                "Verification complete. FW image integrity check passed!")
            success = True

        else:
            self.log.error(f"Verification failed. New state = {state}")
            success = False

        return success

    def upgrade(self):
        """
        Upgrades the FW on the device
        """

        self.log.info("*** FW UPDATE PROCESS STARTED ***")
        # First, check if we need to reset the FSM
        state = self.getState()

        if state is FWState.UNKNOWN:
            self.log.error("No response from probe")
            return 0

        elif state != FWState.IDLE:
            self.log.warning(
                "Probe not ready for FW updates. Resetting the state machine...")
            if self.closeFSM() != 1:
                self.log.error("Unable to reset state machine")
                return 0

        self.log.info("*** Step 1: Enter the probe's bootloader mode ***")
        if self.enterFSM() != 1:
            self.log.error("Could not enter bootloader mod")
            return 0

        self.log.info("*** Step 2: Send number of packets ***")
        if self.sendNumPackets() != 1:
            self.log.error("Could not set number of packets")
            return 0

        self.log.info("*** Step 3: Downloading firmware ***")
        if self.downloadFile() != 1:
            self.log.error("Download failed")
            return 0

        self.log.info("*** Step 4: Setting CRC ***")
        if self.sendImageCRC32() != 1:
            self.log.error("Sending and verifying CRC32 failed")
            return 0

        self.log.info("Waiting for update completion.")
        while self.state not in [FWState.DONE, FWState.ERROR]:
            # TODO: This will advance even when there is an error...
            time.sleep(1)
            self.getState()

        self.log.info("*** Step 5: Apply (close state machine) ***")
        if self.closeFSM() != 1:
            self.log.error("Closing failed")
            return 0

        self.log.info(
            "===>>> FW Update COMPLETE. Device will disconnect and reboot <<<===")

        return 1
